from udsoncan import Request, Response, Baudrate
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseSubfunction, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
import udsoncan.tools as tools

from typing import Optional, cast


class LinkControl(BaseService):
    _sid = 0x87

    supported_negative_response = [ResponseCode.SubFunctionNotSupported,
                                   ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestSequenceError,
                                   ResponseCode.RequestOutOfRange
                                   ]

    class ControlType(BaseSubfunction):
        """
        LinkControl defined subfunctions
        """
        __pretty_name__ = 'control type'

        verifyBaudrateTransitionWithFixedBaudrate = 1
        verifyBaudrateTransitionWithSpecificBaudrate = 2
        transitionBaudrate = 3

    class ResponseData(BaseResponseData):
        """
        .. data:: control_type_echo

                Request subfunction echoed back by the server
        """

        control_type_echo: int

        def __init__(self, control_type_echo: int):
            super().__init__(LinkControl)
            self.control_type_echo = control_type_echo

    class InterpretedResponse(Response):
        service_data: "LinkControl.ResponseData"

    @classmethod
    def make_request(cls, control_type: int, baudrate: Optional[Baudrate] = None) -> Request:
        """
        Generates a request for LinkControl

        :param control_type: Service subfunction. Allowed values are from 0 to 0x7F
        :type control_type: int

        :param baudrate: Required baudrate value when ``control_type`` is either ``verifyBaudrateTransitionWithFixedBaudrate`` (1) or ``verifyBaudrateTransitionWithSpecificBaudrate`` (2)
        :type baudrate: :ref:`Baudrate <Baudrate>`

        :raises ValueError: If parameters are out of range, missing or wrong type
        """

        tools.validate_int(control_type, min=0, max=0x7F, name='Control type')

        if control_type in [cls.ControlType.verifyBaudrateTransitionWithSpecificBaudrate, cls.ControlType.verifyBaudrateTransitionWithFixedBaudrate]:
            if baudrate is None:
                raise ValueError('A Baudrate must be provided with control type : "verifyBaudrateTransitionWithSpecificBaudrate" (0x%02x) or "verifyBaudrateTransitionWithFixedBaudrate" (0x%02x)' % (
                    cls.ControlType.verifyBaudrateTransitionWithSpecificBaudrate, cls.ControlType.verifyBaudrateTransitionWithFixedBaudrate))

            if not isinstance(baudrate, Baudrate):
                raise ValueError('Given baudrate must be an instance of the Baudrate class')
        else:
            if baudrate is not None:
                raise ValueError('The baudrate parameter is only needed when control type is "verifyBaudrateTransitionWithSpecificBaudrate" (0x%02x) or "verifyBaudrateTransitionWithFixedBaudrate" (0x%02x)' % (
                    cls.ControlType.verifyBaudrateTransitionWithSpecificBaudrate, cls.ControlType.verifyBaudrateTransitionWithFixedBaudrate))

        if control_type == cls.ControlType.verifyBaudrateTransitionWithSpecificBaudrate:
            assert baudrate is not None
            baudrate = baudrate.make_new_type(Baudrate.Type.Specific)

        if control_type == cls.ControlType.verifyBaudrateTransitionWithFixedBaudrate:
            assert baudrate is not None
            if baudrate.baudtype == Baudrate.Type.Specific:
                baudrate = baudrate.make_new_type(Baudrate.Type.Fixed)

        request = Request(service=cls, subfunction=control_type)
        if baudrate is not None:
            request.data = baudrate.get_bytes()
        return request

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`LinkControl.ResponseData<udsoncan.services.LinkControl.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If length of ``response.data`` is too short
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 1:
            raise InvalidResponseException(response, "Response data must be at least 1 bytes")

        response.service_data = cls.ResponseData(
            control_type_echo=response.data[0]
        )

        return cast(LinkControl.InterpretedResponse, response)



from udsoncan import services, MemoryLocation, DataFormatIdentifier, DynamicDidDefinition
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest

class TestDynamicallyDefineDataIdentifier(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    # ====== Define by DID ==========

    def test_define_by_did_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x2C\x01\xf2\x01\x12\x34\x01\x02")
        self.conn.fromuserqueue.put(b"\x6c\x01\xf2\x01")    # Positive response

    def _test_define_by_did_success(self):
        response = self.udsclient.dynamically_define_did(0xf201, DynamicDidDefinition(source_did = 0x1234, position=1, memorysize=2))
        self.assertEqual(response.service_data.subfunction_echo, 1)
        self.assertEqual(response.service_data.did_echo, 0xf201)

    def test_define_by_did_composite_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x2C\x01\xf2\x02\x12\x34\x01\x02\x56\x78\x01\x04\x12\x34\x05\x02")
        self.conn.fromuserqueue.put(b"\x6c\x01\xf2\x02")    # Positive response

    def _test_define_by_did_composite_success(self):
        diddef = DynamicDidDefinition(source_did = 0x1234, position=1, memorysize=2)
        diddef.add(source_did = 0x5678, position=1, memorysize=4)
        diddef.add(source_did = 0x1234, position=5, memorysize=2)

        response = self.udsclient.dynamically_define_did(0xf202, diddef)
        self.assertEqual(response.service_data.subfunction_echo, 1)
        self.assertEqual(response.service_data.did_echo, 0xf202)

    def test_define_by_did_composite_wrong_subfunction_echo(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x6c\x02\xf3\x03")    # Positive response

    def _test_define_by_did_composite_wrong_subfunction_echo(self):
        diddef = DynamicDidDefinition(source_did = 0x1234, position=1, memorysize=2)
        diddef.add(source_did = 0x5678, position=1, memorysize=4)
        diddef.add(source_did = 0x1234, position=5, memorysize=2)

        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.dynamically_define_did(0xf203, diddef)

    def test_define_by_did_composite_wrong_did_echo(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x6c\x01\xf3\x05")    # Positive response

    def _test_define_by_did_composite_wrong_did_echo(self):
        diddef = DynamicDidDefinition(source_did = 0x1234, position=1, memorysize=2)
        diddef.add(source_did = 0x5678, position=1, memorysize=4)
        diddef.add(source_did = 0x1234, position=5, memorysize=2)

        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.dynamically_define_did(0xf204, diddef)

    def test_define_by_did_incomplete_response(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x6c") 

    def _test_define_by_did_incomplete_response(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.dynamically_define_did(0xf201, DynamicDidDefinition(source_did = 0x1234, position=1, memorysize=2))




    # ====== Define by Memory Address ==========
    def test_define_by_memloc_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x2C\x02\xf3\x01\x12\x56\x78\x45")
        self.conn.fromuserqueue.put(b"\x6c\x02\xf3\x01")    # Positive response

    def _test_define_by_memloc_success(self):
        memloc = MemoryLocation(address=0x5678, memorysize=0x45 ,address_format=16, memorysize_format=8)
        response = self.udsclient.dynamically_define_did(0xf301, memloc)
        self.assertEqual(response.service_data.subfunction_echo, 2)
        self.assertEqual(response.service_data.did_echo, 0xf301)

    def test_define_by_memloc_composite_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x2C\x02\xf3\x02\x12\x11\x22\x02\x33\x44\x04\x55\x66\x08")
        self.conn.fromuserqueue.put(b"\x6c\x02\xf3\x02")    # Positive response

    def _test_define_by_memloc_composite_success(self):
        diddef = DynamicDidDefinition()
        diddef.add(MemoryLocation(address=0x1122, memorysize=2, address_format=16, memorysize_format=8))
        diddef.add(MemoryLocation(address=0x3344, memorysize=4, address_format=16, memorysize_format=8))
        diddef.add(MemoryLocation(address=0x5566, memorysize=8, address_format=16, memorysize_format=8))

        response = self.udsclient.dynamically_define_did(0xf302, diddef)
        self.assertEqual(response.service_data.subfunction_echo, 2)
        self.assertEqual(response.service_data.did_echo, 0xf302)

    def test_define_by_memloc_composite_success_default_format(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x2C\x02\xf3\x02\x24\x00\x00\x11\x22\x00\x02\x00\x00\x33\x44\x00\x04\x00\x00\x55\x66\x00\x08")
        self.conn.fromuserqueue.put(b"\x6c\x02\xf3\x02")    # Positive response

    def _test_define_by_memloc_composite_success_default_format(self):
        diddef = DynamicDidDefinition()
        diddef.add(MemoryLocation(address=0x1122, memorysize=2))
        diddef.add(MemoryLocation(address=0x3344, memorysize=4))
        diddef.add(MemoryLocation(address=0x5566, memorysize=8))

        self.udsclient.set_config('server_address_format', 32)
        self.udsclient.set_config('server_memorysize_format', 16)

        response = self.udsclient.dynamically_define_did(0xf302, diddef)
        self.assertEqual(response.service_data.subfunction_echo, 2)
        self.assertEqual(response.service_data.did_echo, 0xf302)

    def test_inconsistent_composite_memloc(self):
        pass

    def _test_inconsistent_composite_memloc(self):
        self.udsclient.set_config('server_address_format', 16)
        self.udsclient.set_config('server_memorysize_format', 8)

        diddef = DynamicDidDefinition()
        diddef.add(MemoryLocation(address=0x11, memorysize=2, address_format=8, memorysize_format=8))
        diddef.add(MemoryLocation(address=0x3344, memorysize=4, address_format=16, memorysize_format=8))
        with self.assertRaises(ValueError):
            self.udsclient.dynamically_define_did(0xf300, diddef)

        diddef = DynamicDidDefinition()
        diddef.add(MemoryLocation(address=0x1122, memorysize=2, address_format=16, memorysize_format=8))
        diddef.add(MemoryLocation(address=0x3344, memorysize=4, address_format=16, memorysize_format=16))
        with self.assertRaises(ValueError):
            self.udsclient.dynamically_define_did(0xf300, diddef)

        diddef = DynamicDidDefinition()
        diddef.add(MemoryLocation(address=0x1122, memorysize=2))
        diddef.add(MemoryLocation(address=0x3344, memorysize=4, address_format=32, memorysize_format=8))
        with self.assertRaises(ValueError):
            self.udsclient.dynamically_define_did(0xf300, diddef)

        diddef = DynamicDidDefinition()
        diddef.add(MemoryLocation(address=0x1122, memorysize=2))
        diddef.add(MemoryLocation(address=0x3344, memorysize=4, address_format=16, memorysize_format=16))
        with self.assertRaises(ValueError):
            self.udsclient.dynamically_define_did(0xf300, diddef)

    def test_define_by_memloc_composite_wrong_subfunction_echo(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x6c\x03\xf3\x02")    # Positive response

    def _test_define_by_memloc_composite_wrong_subfunction_echo(self):
        diddef = DynamicDidDefinition()
        diddef.add(MemoryLocation(address=0x1122, memorysize=2, address_format=16, memorysize_format=8))
        diddef.add(MemoryLocation(address=0x3344, memorysize=4, address_format=16, memorysize_format=8))
        diddef.add(MemoryLocation(address=0x5566, memorysize=8, address_format=16, memorysize_format=8))

        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.dynamically_define_did(0xf302, diddef)
 
    def test_define_by_memloc_composite_wrong_did_echo(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x6c\x02\xf3\x03")    # Positive response

    def _test_define_by_memloc_composite_wrong_did_echo(self):
        diddef = DynamicDidDefinition()
        diddef.add(MemoryLocation(address=0x1122, memorysize=2, address_format=16, memorysize_format=8))
        diddef.add(MemoryLocation(address=0x3344, memorysize=4, address_format=16, memorysize_format=8))
        diddef.add(MemoryLocation(address=0x5566, memorysize=8, address_format=16, memorysize_format=8))

        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.dynamically_define_did(0xf302, diddef)       



    # ====== Clear ==========

    def test_clear_dynamic_did(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x2C\x03\xf2\x00")
        self.conn.fromuserqueue.put(b"\x6c\x03\xf2\x00")    # Positive response

    def _test_clear_dynamic_did(self):
        response = self.udsclient.clear_dynamically_defined_did(0xF200)
        self.assertEqual(response.service_data.subfunction_echo, 3)
        self.assertEqual(response.service_data.did_echo,0xf200)

    def test_clear_all_dynamic_did(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x2C\x03")
        self.conn.fromuserqueue.put(b"\x6c\x03")    # Positive response

    def _test_clear_all_dynamic_did(self):
        response = self.udsclient.clear_all_dynamically_defined_did()
        self.assertEqual(response.service_data.subfunction_echo, 3)


    # Error handling

    def test_clear_dynamic_did_wrong_subfunction_echo(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x6c\x02\xf2\x00")    # Positive response

    def _test_clear_dynamic_did_wrong_subfunction_echo(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.clear_dynamically_defined_did(0xF200)

    def test_clear_dynamic_did_wrong_did_echo(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x6c\x03\xf2\x01")    # Positive response

    def _test_clear_dynamic_did_wrong_did_echo(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.clear_dynamically_defined_did(0xF200)


    def test_clear_all_dynamic_did_negative_response(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x2C\x03")
        self.conn.fromuserqueue.put(b"\x7F\x2C\x10")    # General Reject

    def _test_clear_all_dynamic_did_negative_response(self):
        with self.assertRaises(NegativeResponseException) as handle:
            self.udsclient.clear_all_dynamically_defined_did()
        response = handle.exception.response

        self.assertTrue(response.valid)
        self.assertTrue(issubclass(response.service, services.DynamicallyDefineDataIdentifier))
        self.assertEqual(response.code, 0x10)

    def test_clear_all_dynamic_did_negative_response_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x2C\x10")  # General Reject

    def _test_clear_all_dynamic_did_negative_response_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.clear_all_dynamically_defined_did()

        self.assertTrue(response.valid)
        self.assertFalse(response.positive)
        self.assertTrue(issubclass(response.service, services.DynamicallyDefineDataIdentifier))
        self.assertEqual(response.code, 0x10)


    def test_clear_all_dynamic_did_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00\x03")  #Inexistent Service

    def _test_clear_all_dynamic_did_invalidservice_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.clear_all_dynamically_defined_did()

    def test_clear_all_dynamic_did_invalidservice_no_exception(self):
        self.wait_request_and_respond(b"\x00\x03") #Inexistent Service

    def _test_clear_all_dynamic_did_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.clear_all_dynamically_defined_did()
        self.assertFalse(response.valid)

    def test_clear_all_dynamic_did_wrongservice_exception(self):
        self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

    def _test_clear_all_dynamic_did_wrongservice_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.clear_all_dynamically_defined_did()

    def test_clear_all_dynamic_did_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

    def _test_clear_all_dynamic_did_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.clear_all_dynamically_defined_did()
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)


    # ====== Others ==========
    def test_bad_params(self):
        pass

    def _test_bad_params(self):
        empty_diddef = DynamicDidDefinition()
        valid_diddef = DynamicDidDefinition(source_did = 0x1234, position=1, memorysize=1)

        with self.assertRaises(ValueError):
            self.udsclient.clear_dynamically_defined_did(-1)

        with self.assertRaises(ValueError):
            self.udsclient.clear_dynamically_defined_did(0x10000)

        with self.assertRaises(ValueError):
            self.udsclient.clear_dynamically_defined_did('aaa')

        with self.assertRaises(ValueError):
            self.udsclient.dynamically_define_did(-1, valid_diddef)

        with self.assertRaises(ValueError):
            self.udsclient.dynamically_define_did(0x10000, valid_diddef)

        with self.assertRaises(ValueError):
            self.udsclient.dynamically_define_did(0xF200, empty_diddef)

        with self.assertRaises(Exception):
            self.udsclient.dynamically_define_did(0xF200, 'aaa')

        with self.assertRaises(ValueError):
            self.udsclient.dynamically_define_did('aaa', valid_diddef)




from udsoncan import services
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest


class TestControlDTCSettings(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def test_set_on(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x85\x01")
        self.conn.fromuserqueue.put(b"\xC5\x01")  # Positive response

    def _test_set_on(self):
        response = self.udsclient.control_dtc_setting(services.ControlDTCSetting.SettingType.on)
        self.assertEqual(response.service_data.setting_type_echo, services.ControlDTCSetting.SettingType.on)

    def test_set_on_spr(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x85\x81")
        self.conn.fromuserqueue.put("wait")  # Synchronize

    def _test_set_on_spr(self):
        with self.udsclient.suppress_positive_response:
            response = self.udsclient.control_dtc_setting(services.ControlDTCSetting.SettingType.on)
            self.assertEqual(response, None)
        self.conn.fromuserqueue.get(timeout=0.2)  # Avoid closing connection prematurely

    def test_set_on_with_extra_data(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x85\x01\x11\x22\x33")
        self.conn.fromuserqueue.put(b"\xC5\x01")  # Positive response

    def _test_set_on_with_extra_data(self):
        response = self.udsclient.control_dtc_setting(setting_type=services.ControlDTCSetting.SettingType.on, data=b'\x11\x22\x33')
        self.assertEqual(response.service_data.setting_type_echo, services.ControlDTCSetting.SettingType.on)

    def test_set_on_harmless_extra_bytes_in_response(self):
        self.wait_request_and_respond(b"\xC5\x01\x77\x88\x99")  # Positive response

    def _test_set_on_harmless_extra_bytes_in_response(self):
        response = self.udsclient.control_dtc_setting(setting_type=services.ControlDTCSetting.SettingType.on)
        self.assertEqual(response.service_data.setting_type_echo, services.ControlDTCSetting.SettingType.on)

    def test_set_params_denied_exception(self):
        self.wait_request_and_respond(b"\x7F\x85\x45")  # Request Out Of Range

    def _test_set_params_denied_exception(self):
        with self.assertRaises(NegativeResponseException) as handle:
            self.udsclient.control_dtc_setting(setting_type=0x45)

    def test_set_params_denied_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x85\x45")  # Request Out Of Range

    def _test_set_params_denied_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.control_dtc_setting(setting_type=0x45)
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)

    def test_set_params_invalid_service_exception(self):
        self.wait_request_and_respond(b"\x00\x45")  # Inexistent Service

    def _test_set_params_invalid_service_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.control_dtc_setting(setting_type=0x45)

    def test_set_params_invalid_service_no_exception(self):
        self.wait_request_and_respond(b"\x00\x45")  # Inexistent Service

    def _test_set_params_invalid_service_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.control_dtc_setting(setting_type=0x45)
        self.assertFalse(response.valid)

    def test_wrong_service_exception(self):
        self.wait_request_and_respond(b"\x7E\x22")  # Valid but wrong service (Tester Present)

    def _test_wrong_service_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.control_dtc_setting(setting_type=0x22)

    def test_wrong_service_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x22")  # Valid but wrong service (Tester Present)

    def _test_wrong_service_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.control_dtc_setting(setting_type=0x22)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_bad_setting_type_response_exception(self):
        self.wait_request_and_respond(b"\xC5\x23")  # Valid but access type

    def _test_bad_setting_type_response_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.control_dtc_setting(setting_type=0x22)

    def test_bad_setting_type_response_no_exception(self):
        self.wait_request_and_respond(b"\xC5\x23")  # Valid but access type

    def _test_bad_setting_type_response_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.control_dtc_setting(setting_type=0x22)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_bad_param(self):
        pass

    def _test_bad_param(self):
        with self.assertRaises(ValueError):
            self.udsclient.control_dtc_setting(setting_type=-1)

        with self.assertRaises(ValueError):
            self.udsclient.control_dtc_setting(setting_type=0x80)

        with self.assertRaises(ValueError):
            self.udsclient.control_dtc_setting(setting_type=services.ControlDTCSetting.SettingType.on, data=1)

        with self.assertRaises(ValueError):
            self.udsclient.control_dtc_setting(setting_type=services.ControlDTCSetting.SettingType.on, data='asdasdasd')



from udsoncan import services, CommunicationType
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest


class TestCommunicationControl(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def test_comcontrol_enable_node(self):
        for i in range(3):
            request = self.conn.touserqueue.get(timeout=0.2)
            self.assertEqual(request, b"\x28\x00\x01")
            self.conn.fromuserqueue.put(b"\x68\x00")  # Positive response

    def _test_comcontrol_enable_node(self):
        control_type = services.CommunicationControl.ControlType.enableRxAndTx
        com_type = CommunicationType(subnet=CommunicationType.Subnet.node, normal_msg=True)
        response = self.udsclient.communication_control(control_type=control_type, communication_type=com_type)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.control_type_echo, control_type)

        response = self.udsclient.communication_control(control_type=control_type, communication_type=com_type.get_byte())
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.control_type_echo, control_type)

        response = self.udsclient.communication_control(control_type=control_type, communication_type=com_type.get_byte_as_int())
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.control_type_echo, control_type)

    def test_comcontrol_enable_node_spr(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x28\x80\x01")
        self.conn.fromuserqueue.put('wait')  # Synchronize

    def _test_comcontrol_enable_node_spr(self):
        control_type = services.CommunicationControl.ControlType.enableRxAndTx
        com_type = CommunicationType(subnet=CommunicationType.Subnet.node, normal_msg=True)
        with self.udsclient.suppress_positive_response:
            response = self.udsclient.communication_control(control_type=control_type, communication_type=com_type)
            self.assertEqual(response, None)
        self.conn.fromuserqueue.get(timeout=0.2)  # Avoid closing connection prematurely

    def test_comcontrol_disable_subnet(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x28\x03\x33")
        self.conn.fromuserqueue.put(b"\x68\x03")  # Positive response

    def _test_comcontrol_disable_subnet(self):
        control_type = services.CommunicationControl.ControlType.disableRxAndTx
        com_type = CommunicationType(subnet=3, normal_msg=True, network_management_msg=True)
        response = self.udsclient.communication_control(control_type=control_type, communication_type=com_type)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.control_type_echo, control_type)

    def test_comcontrol_negative_response_exception(self):
        self.wait_request_and_respond(b"\x7F\x28\x31") 	# Request Out Of Range

    def _test_comcontrol_negative_response_exception(self):
        with self.assertRaises(NegativeResponseException) as handle:
            control_type = services.CommunicationControl.ControlType.disableRxAndTx
            com_type = CommunicationType(subnet=3, normal_msg=True, network_management_msg=True)
            self.udsclient.communication_control(control_type=control_type, communication_type=com_type)

    def test_comcontrol_negative_response_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x28\x31") 	# Request Out Of Range

    def _test_comcontrol_negative_response_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        control_type = services.CommunicationControl.ControlType.disableRxAndTx
        com_type = CommunicationType(subnet=3, normal_msg=True, network_management_msg=True)
        response = self.udsclient.communication_control(control_type=control_type, communication_type=com_type)
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)

    def test_set_params_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00\x22")  # Inexistent Service

    def _test_set_params_invalidservice_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            control_type = services.CommunicationControl.ControlType.disableRxAndTx
            com_type = CommunicationType(subnet=5, normal_msg=True, network_management_msg=True)
            self.udsclient.communication_control(control_type=control_type, communication_type=com_type)

    def test_set_params_invalidservice_no_exception(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x00\x22")  # Inexistent Service

    def _test_set_params_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        control_type = services.CommunicationControl.ControlType.disableRxAndTx
        com_type = CommunicationType(subnet=5, normal_msg=True, network_management_msg=True)
        response = self.udsclient.communication_control(control_type=control_type, communication_type=com_type)
        self.assertFalse(response.valid)

    def test_comcontrol_wrongservice_exception(self):
        self.wait_request_and_respond(b"\x7E\x22")  # Valid but wrong service (Tester Present)

    def _test_comcontrol_wrongservice_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            control_type = services.CommunicationControl.ControlType.disableRxAndTx
            com_type = CommunicationType(subnet=3, normal_msg=True, network_management_msg=True)
            self.udsclient.communication_control(control_type=control_type, communication_type=com_type)

    def test_comcontrol_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x22")  # Valid but wrong service (Tester Present)

    def _test_comcontrol_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        control_type = services.CommunicationControl.ControlType.disableRxAndTx
        com_type = CommunicationType(subnet=3, normal_msg=True, network_management_msg=True)
        response = self.udsclient.communication_control(control_type=control_type, communication_type=com_type)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_comcontrol_bad_control_type_exception(self):
        self.wait_request_and_respond(b"\x68\x08")  # Valid but bad control type

    def _test_comcontrol_bad_control_type_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            com_type = CommunicationType(subnet=3, normal_msg=True, network_management_msg=True)
            self.udsclient.communication_control(control_type=9, communication_type=com_type)

    def test_comcontrol_bad_control_type_no_exception(self):
        self.wait_request_and_respond(b"\x68\x08")  # Valid but bad control type

    def _test_comcontrol_bad_control_type_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        com_type = CommunicationType(subnet=3, normal_msg=True, network_management_msg=True)
        response = self.udsclient.communication_control(control_type=9, communication_type=com_type)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_bad_param(self):
        pass

    def _test_bad_param(self):
        valid_com_type = CommunicationType(subnet=3, normal_msg=True, network_management_msg=True)
        with self.assertRaises(ValueError):
            self.udsclient.communication_control(control_type='x', communication_type=valid_com_type)

        with self.assertRaises(ValueError):
            self.udsclient.communication_control(control_type=0x80, communication_type=valid_com_type)

        with self.assertRaises(ValueError):
            self.udsclient.communication_control(control_type=-1, communication_type=valid_com_type)

        with self.assertRaises(ValueError):
            self.udsclient.communication_control(control_type=0, communication_type='x')



import struct
from udsoncan import Request, Response
from udsoncan.BaseService import BaseService, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
from udsoncan.exceptions import *
import udsoncan.tools as tools

from typing import Optional, cast


class TransferData(BaseService):
    _sid = 0x36
    _use_subfunction = False

    supported_negative_response = [ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.RequestSequenceError,
                                   ResponseCode.RequestOutOfRange,
                                   ResponseCode.TransferDataSuspended,
                                   ResponseCode.GeneralProgrammingFailure,
                                   ResponseCode.WrongBlockSequenceCounter,
                                   ResponseCode.VoltageTooHigh,
                                   ResponseCode.VoltageTooLow
                                   ]

    class ResponseData(BaseResponseData):
        """
        .. data:: sequence_number_echo

                Requests subfunction echoed back by the server

        .. data:: parameter_records

                Optional additional data associated with the response.
        """

        sequence_number_echo: int
        parameter_records: bytes

        def __init__(self, sequence_number_echo: int, parameter_records: bytes):
            super().__init__(TransferData)
            self.sequence_number_echo = sequence_number_echo
            self.parameter_records = parameter_records

    class InterpretedResponse(Response):
        service_data: "TransferData.ResponseData"

    @classmethod
    def make_request(cls, sequence_number: int, data: Optional[bytes] = None) -> Request:
        """
        Generates a request for TransferData

        :param sequence_number: Corresponds to an 8bit counter that should increment for each new block transferred.
                Allowed values are from 0 to 0xFF
        :type sequence_number: int

        :param data: Optional additional data to send to the server
        :type data: bytes

        :raises ValueError: If parameters are out of range, missing or wrong type
        """

        tools.validate_int(sequence_number, min=0, max=0xFF, name='Block sequence counter')  # Not a subfunction!

        if data is not None and not isinstance(data, bytes):
            raise ValueError('data must be a bytes object')

        request = Request(service=cls)
        request.data = struct.pack('B', sequence_number)

        if data is not None:
            request.data += data
        return request

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`TransferData.ResponseData<udsoncan.services.TransferData.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If length of ``response.data`` is too short
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 1:
            raise InvalidResponseException(response, "Response data must be at least 1 bytes")

        response.service_data = cls.ResponseData(
            sequence_number_echo=response.data[0],
            parameter_records=response.data[1:] if len(response.data) > 1 else bytes()
        )

        return cast(TransferData.InterpretedResponse, response)



from udsoncan.client import Client
from udsoncan import services
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest

class TestTesterPresent(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def test_tester_present_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x3E\x00")
        self.conn.fromuserqueue.put(b"\x7E\x00")	# Positive response

    def _test_tester_present_success(self):
        response = self.udsclient.tester_present()
        self.assertTrue(response.positive)

    def test_tester_present_success_spr(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x3E\x80")
        self.conn.fromuserqueue.put('wait')	# Syncronize

    def _test_tester_present_success_spr(self):
        with self.udsclient.suppress_positive_response:
            response = self.udsclient.tester_present()
            self.assertEqual(response, None)
        self.conn.fromuserqueue.get(timeout=0.2)	#Avoid closing connection prematurely

    def test_tester_present_denied_exception(self):
        self.wait_request_and_respond(b"\x7F\x3E\x13") # IMLOIF

    def _test_tester_present_denied_exception(self):
        with self.assertRaises(NegativeResponseException) as handle:
            self.udsclient.tester_present()
        response = handle.exception.response

        self.assertTrue(response.valid)
        self.assertFalse(response.positive)
        self.assertTrue(issubclass(response.service, services.TesterPresent))
        self.assertEqual(response.code, 0x13)

    def test_tester_present_denied_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x3E\x13") # IMLOIF

    def _test_tester_present_denied_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.tester_present()

        self.assertTrue(response.valid)
        self.assertFalse(response.positive)
        self.assertTrue(issubclass(response.service, services.TesterPresent))
        self.assertEqual(response.code, 0x13)

    def test_tester_present_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00\x00") #Inexistent Service

    def _test_tester_present_invalidservice_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.tester_present()

    def test_tester_present_invalidservice_no_exception(self):
        self.wait_request_and_respond(b"\x00\x00") #Inexistent Service

    def _test_tester_present_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.tester_present()
        self.assertFalse(response.valid)

    def test_tester_present_wrongservice_exception(self):
        self.wait_request_and_respond(b"\x50\x00") # Valid but wrong service (Tester Present)

    def _test_tester_present_wrongservice_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.tester_present()

    def test_tester_present_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x50\x00") # Valid but wrong service (Tester Present)

    def _test_tester_present_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.tester_present()
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)



from test.UdsTest import UdsTest
from test.stub import StubbedIsoTPSocket
from udsoncan.exceptions import *
import socket

class TestStubbedIsoTPSocket(UdsTest):
    def test_open(self):
        tpsock = StubbedIsoTPSocket()
        self.assertFalse(tpsock.bound)
        tpsock.bind(interface='vcan0', rxid=0x100, txid=0x101)
        self.assertTrue(tpsock.bound)
        tpsock.close()
        self.assertFalse(tpsock.bound)

    def test_transmit(self):
        tpsock1 = StubbedIsoTPSocket()
        tpsock2 = StubbedIsoTPSocket(timeout=0.5)
        tpsock1.bind(interface='vcan0', rxid=0x200, txid=0x201)
        tpsock2.bind(interface='vcan0', rxid=0x201, txid=0x200)

        payload1 = b"\x01\x02\x03\x04"
        tpsock1.send(payload1)
        payload2 = tpsock2.recv()
        self.assertEqual(payload1, payload2)

    def test_multicast(self):
        tpsock1 = StubbedIsoTPSocket()
        tpsock2 = StubbedIsoTPSocket(timeout=0.5)
        tpsock3 = StubbedIsoTPSocket(timeout=0.5)
        tpsock1.bind(interface='vcan0', rxid=0x300, txid=0x301)
        tpsock2.bind(interface='vcan0', rxid=0x301, txid=0x300)
        tpsock3.bind(interface='vcan0', rxid=0x301, txid=0x300)

        payload1 = b"\x01\x02\x03\x04"
        tpsock1.send(payload1)
        payload2 = tpsock2.recv()
        payload3 = tpsock3.recv()
        self.assertEqual(payload1, payload2)
        self.assertEqual(payload1, payload3)

    def test_empty_on_close(self):
        tpsock1 = StubbedIsoTPSocket()
        tpsock2 = StubbedIsoTPSocket(timeout=0.2)
        tpsock1.bind(interface='vcan0', rxid=0x400, txid=0x401)
        tpsock2.bind(interface='vcan0', rxid=0x401, txid=0x400)

        payload = b"\x01\x02\x03\x04"
        tpsock1.send(payload)
        tpsock2.close()

        with self.assertRaises(socket.timeout):
            tpsock2.recv()

    def test_no_listener(self):
        tpsock1 = StubbedIsoTPSocket()
        tpsock2 = StubbedIsoTPSocket(timeout=0.2)
        tpsock1.bind(interface='vcan0', rxid=0x400, txid=0x401)

        payload = b"\x01\x02\x03\x04"
        tpsock1.send(payload)
        tpsock2.bind(interface='vcan0', rxid=0x401, txid=0x400)

        with self.assertRaises(socket.timeout):
            tpsock2.recv()




from udsoncan import Request, Response
from udsoncan.BaseService import BaseService, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
from udsoncan.exceptions import *

from typing import Optional, cast


class RequestTransferExit(BaseService):
    _sid = 0x37
    _use_subfunction = False
    _no_response_data = True

    supported_negative_response = [ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.RequestSequenceError
                                   ]

    class ResponseData(BaseResponseData):
        """
        .. data:: parameter_records

                bytes object containing optional data provided by the server
        """

        parameter_records: bytes

        def __init__(self, parameter_records: bytes):
            super().__init__(RequestTransferExit)
            self.parameter_records = parameter_records

    class InterpretedResponse(Response):
        service_data: "RequestTransferExit.ResponseData"

    @classmethod
    def make_request(cls, data: Optional[bytes] = None) -> Request:
        """
        Generates a request for RequestTransferExit

        :param data: Additional optional data to send to the server
        :type data: bytes

        :raises ValueError: If parameters are out of range, missing or wrong type
        """

        if data is not None and not isinstance(data, bytes):
            raise ValueError('data must be a bytes object')

        request = Request(service=cls, data=data)
        return request

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`RequestTransferExit.ResponseData<udsoncan.services.RequestTransferExit.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`
        """
        response.service_data = cls.ResponseData(
            parameter_records=response.data if response.data else bytes()
        )

        return cast(RequestTransferExit.InterpretedResponse, response)



import socket
import queue
import threading
import logging
import binascii
import sys
from abc import ABC, abstractmethod
import time

try:
    import can  # type:ignore
    _import_can_err = None
except Exception as e:
    _import_can_err = e

try:
    import isotp    # type:ignore
    _import_isotp_err = None
except Exception as e:
    _import_isotp_err = e

try:
    from udsoncan.j2534 import J2534
    from udsoncan.j2534 import Protocol_ID, Ioctl_Flags, Ioctl_ID
    _import_j2534_err = None
except Exception as e:
    _import_j2534_err = e

try:
    from aioisotp.sync import SyncISOTPNetwork  # type:ignore
    _import_aioisotp_err = None
except Exception as e:
    _import_aioisotp_err = e

from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan.exceptions import TimeoutException


class BaseConnection(ABC):

    def __init__(self, name=None):
        if name is None:
            self.name = 'Connection'
        else:
            self.name = 'Connection[%s]' % (name)

        self.logger = logging.getLogger(self.name)

    def send(self, data):
        """Sends data to the underlying transport protocol

        :param data: The data or object to send. If a Request or Response is given, the value returned by get_payload() will be sent.
        :type data: bytes, Request, Response

        :returns: None
        """

        if isinstance(data, Request) or isinstance(data, Response):
            payload = data.get_payload()
        else:
            payload = data

        self.logger.debug('Sending %d bytes : [%s]' % (len(payload), binascii.hexlify(payload)))
        self.specific_send(payload)

    def wait_frame(self, timeout=2, exception=False):
        """Waits for the reception of a frame of data from the underlying transport protocol

        :param timeout: The maximum amount of time to wait before giving up in seconds
        :type timeout: int
        :param exception: Boolean value indicating if this function may return exceptions.
                When ``True``, all exceptions may be raised, including ``TimeoutException``
                When ``False``, all exceptions will be logged as ``DEBUG`` and ``None`` will be returned.
        :type exception: bool

        :returns: Received data
        :rtype: bytes or None
        """
        try:
            frame = self.specific_wait_frame(timeout=timeout)
        except Exception as e:
            self.logger.debug('No data received: [%s] - %s ' % (e.__class__.__name__, str(e)))

            if exception == True:
                raise
            else:
                frame = None

        if frame is not None:
            self.logger.debug('Received %d bytes : [%s]' % (len(frame), binascii.hexlify(frame)))
        return frame

    def __enter__(self):
        return self

    @abstractmethod
    def specific_send(self, payload):
        """The implementation of the send method.

        :param payload: Data to send
        :type payload: bytes

        :returns: None
        """
        pass

    @abstractmethod
    def specific_wait_frame(self, timeout=2):
        """The implementation of the ``wait_frame`` method. 

        :param timeout: The maximum amount of time to wait before giving up
        :type timeout: int

        :returns: Received data
        :rtype: bytes or None
        """
        pass

    @abstractmethod
    def open(self):
        """ Set up the connection object. 

        :returns: None
        """
        pass

    @abstractmethod
    def close(self):
        """ Close the connection object

        :returns: None
        """
        pass

    @abstractmethod
    def empty_rxqueue(self):
        """ Empty all unread data in the reception buffer.

        :returns: None
        """
        pass

    @abstractmethod
    def is_open(self) -> bool:
        """ Tells if the connection is open.

        :returns: bool
        """
        pass

    def __exit__(self, type, value, traceback):
        pass


class SocketConnection(BaseConnection):
    """
    Sends and receives data through a socket.

    :param sock: The socket to use. This socket must be bound and ready to use. Only ``send()`` and ``recv()`` will be called by this Connection
    :type sock: socket.socket
    :param bufsize: Maximum buffer size of the socket, this value is passed to ``recv()``
    :type bufsize: int
    :param name: This name is included in the logger name so that its output can be redirected. The logger name will be ``Connection[<name>]``
    :type name: string

    """

    def __init__(self, sock, bufsize=4095, name=None):
        BaseConnection.__init__(self, name)

        self.rxqueue = queue.Queue()
        self.exit_requested = False
        self.opened = False
        self.rxthread = None
        self.sock = sock
        self.sock.settimeout(0.1)  # for recv
        self.bufsize = bufsize

    def open(self):
        self.exit_requested = False
        self.rxthread = threading.Thread(target=self.rxthread_task, daemon=True)
        self.rxthread.start()
        self.opened = True
        self.logger.info('Connection opened')
        return self

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def is_open(self):
        return self.opened

    def rxthread_task(self):
        while not self.exit_requested:
            try:
                data = self.sock.recv(self.bufsize)
                if data is not None:
                    self.rxqueue.put(data)
            except socket.timeout:
                pass
            except Exception:
                self.exit_requested = True

    def close(self):
        self.exit_requested = True
        self.rxthread.join()
        self.opened = False
        self.logger.info('Connection closed')

    def specific_send(self, payload):
        self.sock.send(payload)

    def specific_wait_frame(self, timeout=2):
        if not self.opened:
            raise RuntimeError("Connection is not open")

        timedout = False
        frame = None
        try:
            frame = self.rxqueue.get(block=True, timeout=timeout)

        except queue.Empty:
            timedout = True

        if timedout:
            raise TimeoutException("Did not received frame in time (timeout=%s sec)" % timeout)

        return frame

    def empty_rxqueue(self):
        while not self.rxqueue.empty():
            self.rxqueue.get()


class IsoTPSocketConnection(BaseConnection):
    """
    Sends and receives data through an ISO-TP socket. Makes cleaner code than SocketConnection but offers no additional functionality.
    The `can-isotp module <https://github.com/pylessard/python-can-isotp>`_ must be installed in order to use this connection

    :param interface: The can interface to use (example: `can0`)
    :type interface: string
    :param rxid: The reception CAN id
    :type rxid: int 
    :param txid: The transmission CAN id
    :type txid: int
    :param name: This name is included in the logger name so that its output can be redirected. The logger name will be ``Connection[<name>]``
    :type name: string
    :param tpsock: An optional ISO-TP socket to use instead of creating one.
    :type tpsock: isotp.socket
    :param args: Optional parameters list passed to ISO-TP socket binding method.
    :type args: list
    :param kwargs: Optional parameters dictionary passed to ISO-TP socket binding method.
    :type kwargs: dict

    """

    def __init__(self, interface, rxid, txid, name=None, tpsock=None, *args, **kwargs):

        BaseConnection.__init__(self, name)

        self.interface = interface
        self.rxid = rxid
        self.txid = txid
        self.rxqueue = queue.Queue()
        self.exit_requested = False
        self.opened = False
        self.tpsock_bind_args = args
        self.tpsock_bind_kwargs = kwargs

        if tpsock is None:
            if 'isotp' not in sys.modules:
                if _import_isotp_err is None:
                    raise ImportError('isotp module is not loaded')
                else:
                    raise _import_isotp_err
            self.tpsock = isotp.socket(timeout=0.1)
        else:
            self.tpsock = tpsock

    def open(self):
        self.tpsock.bind(self.interface, rxid=self.rxid, txid=self.txid, *self.tpsock_bind_args, **self.tpsock_bind_kwargs)
        self.exit_requested = False
        self.rxthread = threading.Thread(target=self.rxthread_task, daemon=True)
        self.rxthread.start()
        self.opened = True
        self.logger.info('Connection opened')
        return self

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def is_open(self):
        return self.tpsock.bound

    def rxthread_task(self):
        while not self.exit_requested:
            try:
                data = self.tpsock.recv()
                if data is not None:
                    self.rxqueue.put(data)
            except socket.timeout:
                pass
            except Exception:
                self.exit_requested = True

    def close(self):
        self.exit_requested = True
        self.rxthread.join()
        self.tpsock.close()
        self.opened = False
        self.logger.info('Connection closed')

    def specific_send(self, payload):
        self.tpsock.send(payload)

    def specific_wait_frame(self, timeout=2):
        if not self.opened:
            raise RuntimeError("Connection is not open")

        timedout = False
        frame = None
        try:
            frame = self.rxqueue.get(block=True, timeout=timeout)

        except queue.Empty:
            timedout = True

        if timedout:
            raise TimeoutException("Did not received ISOTP frame in time (timeout=%s sec)" % timeout)

        return frame

    def empty_rxqueue(self):
        while not self.rxqueue.empty():
            self.rxqueue.get()


class IsoTPConnection(IsoTPSocketConnection):
    """
    Same as :class:`IsoTPSocketConnection <udsoncan.connections.IsoTPSocketConnection.Session>`. Exists only for backward compatibility. 
    """
    pass


class QueueConnection(BaseConnection):
    """
    Sends and receives data using 2 Python native queues.

    - ``MyConnection.fromuserqueue`` : Data read from this queue when ``wait_frame`` is called
    - ``MyConnection.touserqueue`` : Data written to this queue when ``send`` is called

    :param mtu: Optional maximum frame size. Messages will be truncated to this size
    :type mtu: int
    :param name: This name is included in the logger name so that its output can be redirected. The logger name will be ``Connection[<name>]``
    :type name: string

    """

    def __init__(self, name=None, mtu=4095):
        BaseConnection.__init__(self, name)

        self.fromuserqueue = queue.Queue()  # Client reads from this queue. Other end is simulated
        self.touserqueue = queue.Queue()  # Client writes to this queue. Other end is simulated
        self.opened = False
        self.mtu = mtu

    def open(self):
        self.opened = True
        self.logger.info('Connection opened')
        return self

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def is_open(self):
        return self.opened

    def close(self):
        self.empty_rxqueue()
        self.empty_txqueue()
        self.opened = False
        self.logger.info('Connection closed')

    def specific_send(self, payload):
        if self.mtu is not None:
            if len(payload) > self.mtu:
                self.logger.warning("Truncating payload to be set to a length of %d" % (self.mtu))
                payload = payload[0:self.mtu]

        self.touserqueue.put(payload)

    def specific_wait_frame(self, timeout=2):
        if not self.opened:
            raise RuntimeError("Connection is not open")

        timedout = False
        frame = None
        try:
            frame = self.fromuserqueue.get(block=True, timeout=timeout)
        except queue.Empty:
            timedout = True

        if timedout:
            raise TimeoutException("Did not receive frame from user queue in time (timeout=%s sec)" % timeout)

        if self.mtu is not None:
            if frame is not None and len(frame) > self.mtu:
                self.logger.warning("Truncating received payload to a length of %d" % (self.mtu))
                frame = frame[0:self.mtu]

        return frame

    def empty_rxqueue(self):
        while not self.fromuserqueue.empty():
            self.fromuserqueue.get()

    def empty_txqueue(self):
        while not self.touserqueue.empty():
            self.touserqueue.get()


class PythonIsoTpConnection(BaseConnection):
    """
    Sends and receives data using a `can-isotp <https://github.com/pylessard/python-can-isotp>`_ Python module which is a Python implementation of the IsoTp transport protocol
    which can be coupled with `python-can <https://python-can.readthedocs.io>`_ module to interract with CAN hardware

    `can-isotp <https://github.com/pylessard/python-can-isotp>`_ must be installed in order to use this connection.

    See an :ref:`example<example_using_python_can>`

    :param isotp_layer: The IsoTP Transport layer object coming from the ``isotp`` module.
    :type isotp_layer: :class:`isotp.TransportLayer<isotp.TransportLayer>`

    :param name: This name is included in the logger name so that its output can be redirected. The logger name will be ``Connection[<name>]``
    :type name: string

    """
    mtu = 4095

    def __init__(self, isotp_layer, name=None):
        BaseConnection.__init__(self, name)
        self.toIsoTPQueue = queue.Queue()
        self.fromIsoTPQueue = queue.Queue()
        self.rxthread = None
        self.exit_requested = False
        self.opened = False
        self.isotp_layer = isotp_layer

        assert isinstance(self.isotp_layer, isotp.TransportLayer), 'isotp_layer must be a valid isotp.TransportLayer '

    def open(self, bus=None):
        if bus is not None:
            self.isotp_layer.set_bus(bus)

        self.exit_requested = False
        self.rxthread = threading.Thread(target=self.rxthread_task, daemon=True)
        self.rxthread.start()
        self.opened = True
        self.logger.info('Connection opened')
        return self

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def is_open(self):
        return self.opened

    def close(self):
        self.empty_rxqueue()
        self.empty_txqueue()
        self.exit_requested = True
        self.rxthread.join()
        self.isotp_layer.reset()
        self.opened = False
        self.logger.info('Connection closed')

    def specific_send(self, payload):
        if self.mtu is not None:
            if len(payload) > self.mtu:
                self.logger.warning("Truncating payload to be set to a length of %d" % (self.mtu))
                payload = payload[0:self.mtu]

        self.toIsoTPQueue.put(bytearray(payload))  # isotp.protocol.TransportLayer uses byte array. udsoncan is strict on bytes format

    def specific_wait_frame(self, timeout=2):
        if not self.opened:
            raise RuntimeError("Connection is not open")

        timedout = False
        frame = None
        try:
            frame = self.fromIsoTPQueue.get(block=True, timeout=timeout)
        except queue.Empty:
            timedout = True

        if timedout:
            raise TimeoutException("Did not receive frame IsoTP Transport layer in time (timeout=%s sec)" % timeout)

        if self.mtu is not None:
            if frame is not None and len(frame) > self.mtu:
                self.logger.warning("Truncating received payload to a length of %d" % (self.mtu))
                frame = frame[0:self.mtu]

        return bytes(frame)  # isotp.protocol.TransportLayer uses bytearray. udsoncan is strict on bytes format

    def empty_rxqueue(self):
        while not self.fromIsoTPQueue.empty():
            self.fromIsoTPQueue.get()

    def empty_txqueue(self):
        while not self.toIsoTPQueue.empty():
            self.toIsoTPQueue.get()

    def rxthread_task(self):
        while not self.exit_requested:
            try:
                while not self.toIsoTPQueue.empty():
                    self.isotp_layer.send(self.toIsoTPQueue.get())

                self.isotp_layer.process()

                while self.isotp_layer.available():
                    self.fromIsoTPQueue.put(self.isotp_layer.recv())

                time.sleep(self.isotp_layer.sleep_time())

            except Exception as e:
                self.exit_requested = True
                self.logger.error(str(e))


class J2534Connection(BaseConnection):
    """
    Sends and receives data through a J2534 Interface. 
    A windows DLL and a J2534 interface must be installed in order to use this connection

    :param windll: The path to the windows DLL for the J2534 interface (example: 'C:/Program Files{x86}../../openport 2.0/op20pt32.dll')
    :type interface: string
    :param rxid: The reception CAN id
    :type rxid: int 
    :param txid: The transmission CAN id
    :type txid: int
    :param name: This name is included in the logger name so that its output can be redirected. The logger name will be ``Connection[<name>]``
    :type name: string
    :param debug: This will enable windows debugging mode in the dll (see tactrix doc for additional information)
    :type debug: boolean
    :param args: Optional parameters list (Unused right now).
    :type args: list
    :param kwargs: Optional parameters dictionary Unused right now).
    :type kwargs: dict

    """

    def __init__(self, windll, rxid, txid, name=None, debug=False, *args, **kwargs):

        BaseConnection.__init__(self, name)

        # Set up a J2534 interface using the DLL provided
        self.interface = J2534(windll=windll, rxid=rxid, txid=txid)

        # Set the protocol to ISO15765, Baud rate to 500000
        self.protocol = Protocol_ID.ISO15765
        self.baudrate = 500000

        # Open the interface (connect to the DLL)
        result, self.devID = self.interface.PassThruOpen()

        if debug:
            result = self.interface.PassThruIoctl(Handle=0, IoctlID=Ioctl_Flags.TX_IOCTL_SET_DLL_DEBUG_FLAGS,
                                                  ioctlInput=Ioctl_Flags.TX_IOCTL_DLL_DEBUG_FLAG_J2534_CALLS)

        # Get the firmeware and DLL version etc, mainly for debugging output
        self.result, self.firmwareVersion, self.dllVersion, self.apiVersion = self.interface.PassThruReadVersion(self.devID)
        self.logger.info("J2534 FirmwareVersion: " + str(self.firmwareVersion.value) + ", dllVersoin: " +
                         str(self.dllVersion.value) + ", apiVersion" + str(self.apiVersion.value))

        # get the channel ID of the interface (used for subsequent communication)
        self.result, self.channelID = self.interface.PassThruConnect(self.devID, self.protocol.value, self.baudrate)

        # Set the filters and clear the read buffer (filters will be set based on tx/rxids)
        self.result = self.interface.PassThruStartMsgFilter(self.channelID, self.protocol.value)
        self.result = self.interface.PassThruIoctl(self.channelID, Ioctl_ID.CLEAR_RX_BUFFER)

        self.rxqueue = queue.Queue()
        self.exit_requested = False
        self.opened = False

    def open(self):
        self.exit_requested = False
        self.rxthread = threading.Thread(target=self.rxthread_task, daemon=True)
        self.rxthread.start()
        self.opened = True
        self.logger.info('J2534 Connection opened')
        return self

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def is_open(self):
        return self.opened

    def rxthread_task(self):

        while not self.exit_requested:

            try:
                result, data, numMessages = self.interface.PassThruReadMsgs(self.channelID, self.protocol.value, 1, 1)

                if data is not None:
                    self.rxqueue.put(data)
            except Exception:
                self.logger.critical("Exiting J2534 rx thread")
                self.exit_requested = True

    def close(self):
        self.exit_requested = True
        self.rxthread.join()
        result = self.interface.PassThruDisconnect(self.channelID)
        self.opened = False
        self.logger.info('J2534 Connection closed')

    def specific_send(self, payload):
        result = self.interface.PassThruWriteMsgs(self.channelID, payload, self.protocol.value)

    def specific_wait_frame(self, timeout=4):
        if not self.opened:
            raise RuntimeError("J2534 Connection is not open")

        timedout = False
        frame = None
        try:
            frame = self.rxqueue.get(block=True, timeout=timeout)

        except queue.Empty:
            timedout = True

        if timedout:
            raise TimeoutException("Did not received response from J2534 RxQueue (timeout=%s sec)" % timeout)

        return frame

    def empty_rxqueue(self):
        while not self.rxqueue.empty():
            self.rxqueue.get()


class FakeConnection(BaseConnection):
    """
    Sends and receives STATIC data defined in a local dict. 
    Used so that an application can be tested without a live can network
    """

    def __init__(self, name=None, debug=False, *args, **kwargs):

        BaseConnection.__init__(self, name)

        self.rxqueue = queue.Queue()

        self.exit_requested = False
        self.opened = False

        self.ResponseData = {b'\x10\x03': b'\x50\x03\x12\x23\x34\x45',
                             b'\x22\xf1\x90\xf1\x89\xf1\x91\xf8\x06\xf1\xa3': b'\x22\xf1\x90\xf1\x89\xf1\x91\xf8\x06\xf1\xa3'}

    def open(self):
        self.opened = True
        self.logger.info('Fake Connection opened')
        return self

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def is_open(self):
        return self.opened

    def close(self):
        self.exit_requested = True
        self.opened = False
        self.logger.info('Fake Connection closed')

    def specific_send(self, payload):
        self.rxqueue.put(self.ResponseData[payload])

    def specific_wait_frame(self, timeout=4):
        if not self.opened:
            raise RuntimeError("Fake Connection is not open")

        timedout = False
        frame = None
        try:
            frame = self.rxqueue.get(block=True, timeout=timeout)
            #frame = self.rxqueue.get(block=True, timeout=5)

        except queue.Empty:
            timedout = True

        if timedout:
            raise TimeoutException("Did not received response from J2534 RxQueue (timeout=%s sec)" % timeout)

        return frame

    def empty_rxqueue(self):
        while not self.rxqueue.empty():
            self.rxqueue.get()


class SyncAioIsotpConnection(BaseConnection):
    """
    A wrapper for aioisotp sync variant

    `aioisotp <https://github.com/christiansandberg/aioisotp>`_ must be installed in order to use this connection.

    See an :ref:`example<example_using_aioisotp>`

    :param rxid: The reception CAN id
    :type rxid: int

    :param txid: The transmission CAN id
    :type txid: int

    :param name: This name is included in the logger name so that its output can be redirected. The logger name will be ``Connection[<name>]``
    :type name: string

    :param args: Optional parameters list passed to aioisotp binding method.
    :type args: list

    :param kwargs: Optional parameters dictionary passed to aioisotp binding method.
    :type kwargs: dict
    """

    def __init__(self, rx_id, tx_id, name=None, *args, **kwargs):
        BaseConnection.__init__(self, name)
        self.network = SyncISOTPNetwork(*args, **kwargs)
        self.opened = False
        self.rx_id = rx_id
        self.tx_id = tx_id

    def specific_send(self, payload):
        self.conn.send(payload)

    def specific_wait_frame(self, timeout=2):
        if not self.opened:
            raise RuntimeError("Connection is not open")

        frame = self.conn.recv(timeout)

        if frame is None and timeout:
            raise TimeoutException("Did not received frame in time (timeout=%s sec)" % timeout)

        return frame

    def open(self):
        self.network.open()
        self.conn = self.network.create_sync_connection(self.rx_id, self.tx_id)
        self.opened = True
        self.logger.info("Connection opened")
        return self

    def close(self):
        self.network.close()
        self.opened = False
        self.logger.info("Connection closed")

    def empty_rxqueue(self):
        self.conn.empty()

    def is_open(self):
        return self.opened

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='udsoncan',
    packages=find_packages(exclude=['test']),
    package_data={
        '': ['*.conf'],
    },
    version='1.17',
    description='Implementation of the Unified Diagnostic Service (UDS) protocol (ISO-14229) used in the automotive industry.',
    long_description=long_description,
    author='Pier-Yves Lessard',
    author_email='py.lessard@gmail.com',
    license='MIT',
    url='https://github.com/pylessard/python-udsoncan',
    download_url='https://github.com/pylessard/python-udsoncan/archive/v1.17.tar.gz',
    keywords=['uds', '14229', 'iso-14229', 'diagnostic', 'automotive'],
    python_requires='>=3.0',
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX :: Linux",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
    ],
)



from udsoncan import services
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest


class TestDiagnosticSessionControl(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def test_dsc_success_2006(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x10\x01")
        self.conn.fromuserqueue.put(b"\x50\x01\x99\x88")    # Positive response

    def _test_dsc_success_2006(self):
        self.udsclient.set_config('standard_version', 2006)
        response = self.udsclient.change_session(services.DiagnosticSessionControl.Session.defaultSession)
        self.assertEqual(response.service_data.session_echo, 1)
        self.assertEqual(response.service_data.session_param_records, b"\x99\x88")
        self.assertIsNone(response.service_data.p2_server_max)
        self.assertIsNone(response.service_data.p2_star_server_max)

    def test_dsc_success_2013_plus(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x10\x01")
        self.conn.fromuserqueue.put(b"\x50\x01\x99\x88\x12\x34")    # Positive response

    def _test_dsc_success_2013_plus(self):
        self.udsclient.set_config('standard_version', 2013)
        self.udsclient.set_config('use_server_timing', True)
        response = self.udsclient.change_session(services.DiagnosticSessionControl.Session.defaultSession)
        self.assertEqual(response.service_data.session_echo, 1)
        self.assertEqual(response.service_data.session_param_records, b"\x99\x88\x12\x34")
        self.assertEqual(response.service_data.p2_server_max, (0x9988) / 1000)
        self.assertEqual(response.service_data.p2_star_server_max, 0x1234 * 10 / 1000)
        self.assertEqual(self.udsclient.session_timing['p2_server_max'], response.service_data.p2_server_max)
        self.assertEqual(self.udsclient.session_timing['p2_star_server_max'], response.service_data.p2_star_server_max)

    def test_dsc_success_2013_plus_ignore_server_timing(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x10\x01")
        self.conn.fromuserqueue.put(b"\x50\x01\x99\x88\x12\x34")    # Positive response

    def _test_dsc_success_2013_plus_ignore_server_timing(self):
        self.udsclient.set_config('standard_version', 2013)
        self.udsclient.set_config('use_server_timing', False)
        response = self.udsclient.change_session(services.DiagnosticSessionControl.Session.defaultSession)
        self.assertEqual(response.service_data.session_echo, 1)
        self.assertEqual(response.service_data.session_param_records, b"\x99\x88\x12\x34")
        self.assertEqual(response.service_data.p2_server_max, 0x9988 / 1000)
        self.assertEqual(response.service_data.p2_star_server_max, 0x1234 * 10 / 1000)
        self.assertIsNone(self.udsclient.session_timing['p2_server_max'])
        self.assertIsNone(self.udsclient.session_timing['p2_star_server_max'])

    def test_dsc_success_spr(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x10\x81")
        self.conn.fromuserqueue.put("wait")  # Synchronize

    def _test_dsc_success_spr(self):
        with self.udsclient.suppress_positive_response:
            response = self.udsclient.change_session(services.DiagnosticSessionControl.Session.defaultSession)
            self.assertEqual(response, None)
        self.conn.fromuserqueue.get(timeout=0.2)  # Avoid closing connection prematurely

    def test_dsc_denied_exception(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x10\x08")
        self.conn.fromuserqueue.put(b"\x7F\x10\x12")  # Subfunction not supported

    def _test_dsc_denied_exception(self):
        with self.assertRaises(NegativeResponseException) as handle:
            self.udsclient.change_session(0x08)
        response = handle.exception.response

        self.assertTrue(response.valid)
        self.assertTrue(issubclass(response.service, services.DiagnosticSessionControl))
        self.assertEqual(response.code, 0x12)

    def test_dsc_denied_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x10\x12")  # Subfunction not supported

    def _test_dsc_denied_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.change_session(0x08)
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)

    def test_dsc_bad_subfunction_exception(self):
        self.wait_request_and_respond(b"\x50\x02\x11\x11\x11\x11")  # Positive response

    def _test_dsc_bad_subfunction_exception(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.change_session(services.DiagnosticSessionControl.Session.defaultSession)

    def test_dsc_bad_subfunction_no_exception(self):
        self.wait_request_and_respond(b"\x50\x02\x11\x11\x11\x11")  # Positive response

    def _test_dsc_bad_subfunction_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.change_session(services.DiagnosticSessionControl.Session.defaultSession)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_dsc_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00\x02\x11\x11\x11\x11")  # Inexistent Service

    def _test_dsc_invalidservice_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.change_session(0x02)

    def test_dsc_invalidservice_no_exception(self):
        self.wait_request_and_respond(b"\x00\x02\x11\x11\x11\x11")  # Inexistent Service

    def _test_dsc_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.change_session(0x02)
        self.assertFalse(response.valid)

    def test_ecu_reset_wrongservice_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        self.wait_request_and_respond(b"\x7E\x00")  # Valid but wrong service (Tester Present)

    def _test_ecu_reset_wrongservice_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.change_session(0x55)

    def test_ecu_reset_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x00")  # Valid but wrong service (Tester Present)

    def _test_ecu_reset_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.change_session(0x55)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_bad_param(self):
        pass

    def _test_bad_param(self):
        with self.assertRaises(ValueError):
            success = self.udsclient.change_session(0x100)

        with self.assertRaises(ValueError):
            success = self.udsclient.change_session(-1)
